/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 539 $ $LastChangedDate: 2002-08-25 13:25:13 +0200 (Sun, 25 Aug 2002) $
 *
 * Codegen - generate Ksi code from Gont Abstract Syntax Trees. 
 */

/*
I use -kr -i8 for C, but -kr -i4 for Gont. This is motivated by:
  - nested functions, that allow tree-like code organization impossible
    in C, i.e. functions are small *not counting* nested functions
  - generally longer lines, mainly because of union constructors
*/

open Ast;
open Util;
open Type;

open Codegen;

struct cg_ctx {
    // cf_ == current_function_
    string cf_name;
    bool cf_is_public;
    ty cf_ret_type_instanced;
    string cf_closure_type_name;
    <cg_ctx>option parent_func;
    <cg_ctx>option parent;
   
    // we cannot use just <var_symbol, Ksi::expr>Ptrhash::t here, and
    // Ptrhash::iter() later to make a list, since Ptrhash::iter() works
    // in random order, which breaks bootstrap comparision
    <Ksi::expr>List::t cf_references;
    <var_symbol, int>Ptrhash::t cf_references_hash;
}

Stdio::FILE ksi_sink;

<cg_ctx>option global_ctx_o;

<Ksi::id>List::t module_init;
<Ksi::id>List::t module_fini;
<Ksi::expr>List::t static_inits;

section init 
{
    global_ctx_o = None;
    module_init = List::Nil;
    module_fini = List::Nil;
    static_inits = List::Nil;
}

void init(Stdio::FILE ksi_file)
{
    ksi_sink = ksi_file;
    
    global_ctx_o = Some[({cf_name = "?global_scope?",
			  cf_is_public = true,
			  cf_closure_type_name = "",
			  cf_ret_type_instanced = AbstractType,
			  cf_references = List::Nil,
			  cf_references_hash = Ptrhash::create(13),
			  parent = None,
			  parent_func = None
			 })];
    Ksi::output_header(ksi_sink);
    Ksi::output_decl(ksi_sink, Ksi::InputFile[(top_location()).file]);
    push_basic_types();
}

cg_ctx get_global_ctx()
{
    switch global_ctx_o {
    case Some[c]:
	return c;
	
    case None:
	return ice("Codegen::get_global_ctx(): not inialized");
    }
}

void output_decl(Ksi::decl d)
{
    Ksi::output_decl(ksi_sink, d);
}

void top_comment(string s)
{
    output_decl(Ksi::TopComment[s]);
}

Ksi::ty gt_int;
Ksi::ty gt_void;
Ksi::ty gt_tvoid;
Ksi::ty gt_bool;
Ksi::ty gt_tbool;
Ksi::ty gt_float;
Ksi::ty gt_tfloat;
Ksi::ty gt_string;
Ksi::ty gt_exn;

Ksi::ty ptroff;
Ksi::ty ksi_void;
Ksi::ty c_double;
Ksi::ty c_double_ptr;
Ksi::ty gont_string;
Ksi::ty gont_lambda_expr;
Ksi::ty gont_union;
Ksi::ty bool_t;
Ksi::ty poly_ptr;
Ksi::ty poly;
Ksi::id our_closure_id;
Ksi::id parent_closure_id;
Ksi::expr our_closure_ref;
Ksi::expr parent_closure_ref;
Ksi::id dummy_id;
Ksi::id union_tag_id;
Ksi::id union_value_id;
Ksi::expr null_value;
Ksi::ty null_type;
Ksi::expr nop;
Ksi::expr true_const;
Ksi::expr false_const;
Ksi::expr zero_const;
Ksi::expr one_const;

<Ksi::id>Hashtbl::t mangled_types;
int mangled_id;
int mangled_struct_id;

Ksi::id make_id(string s)
{
    return Ksi::make_id(s);
}

section init
{
    our_closure_id = make_id("closure$");
    parent_closure_id = make_id("parent_closure$");
    our_closure_ref = Ksi::Ref[our_closure_id];
    parent_closure_ref = Ksi::Ref[parent_closure_id];
    dummy_id = make_id("__dummy$__");
    null_type = Ksi::Ptr[Ksi::Void];
    null_value = Ksi::Null[null_type];
    nop = Ksi::Nop;
    ptroff = Ksi::MacroType[make_id("ptr_off")];
    c_double = Ksi::MacroType[make_id("c_double")];
    c_double_ptr = Ksi::Ptr[c_double];
    gont_string = Ksi::Ptr[Ksi::Struct[make_id("gont_string")]];
    gont_lambda_expr = Ksi::Ptr[Ksi::Struct[make_id("gont_lambda_expr")]];
    gont_union = Ksi::Ptr[Ksi::Struct[make_id("gont_union")]];
    bool_t = Ksi::Bool;
    poly = Ksi::Poly;
    ksi_void = Ksi::Void;
    poly_ptr = Ksi::Ptr[poly];

    true_const = Ksi::BoolConst[true];
    false_const = Ksi::BoolConst[false];
    zero_const = Ksi::IntConst[ptroff, Int64::zero];
    one_const = Ksi::IntConst[ptroff, Int64::one];
    
    union_tag_id = make_id("tag");
    union_value_id = make_id("value");
    
    mangled_types = Hashtbl::create(203);
    mangled_id = 1;
    mangled_struct_id = 1;
}

Ksi::ty flat_type(ty t, bool storage)
{
    Ksi::ty push_struct(<<sf_symbol>field>List::t lst)
    {
	*[Ksi::ty, Ksi::id] make_field(<sf_symbol>field f)
	{
	    let s = get_binding(f.fld_symbol) in
	    return [flat_type(f.fld_type, true), make_id(s.name)];
	}

	let m = "s:" + mangle_type(Struct[lst]) in
	if (Hashtbl::mem(mangled_types, m)) {
	    return Ksi::Ptr[Ksi::Struct[Hashtbl::find(mangled_types, m)]];
	} else {
	    Ksi::id n = make_id("s" + itoa(mangled_struct_id++));
	    Hashtbl::add(mangled_types, m, n);
	    let c = List::map(make_field, lst) in {
	       output_decl(Ksi::StructDef[List::Nil, n, c]);
	       return Ksi::Ptr[Ksi::Struct[n]];
	    }
	}
    }
    
    switch type_instance(t) {
    case Int:			return ptroff;
    case String:		return gont_string;
    case Tuple[_]:		return poly_ptr;
    case Func[_]:		return gont_lambda_expr;
    case TyVar[_]:		return poly;
    case XTyVar[_]:		return poly;
    case Struct[lst]:		return push_struct(lst);
    case OptStruct[lst]:	return push_struct(lst);
    case Union[_]:		return gont_union;
    case AbstractType[_]:	return ptroff;
    case Exn:			return gont_union;
    case NamedType[_]: 
	return ice("Codegen::flat_type(): NamedType");
    case Bool:
	if (storage)    return bool_t;
	else		return ptroff;
    case Float:	
	if (storage)    return c_double;
	else		return c_double_ptr;
    case Void:			
	if (storage)	return ptroff;
	else		return ksi_void;
    }
}

void push_basic_types()
{
    Ksi::ty ini(string name, Ksi::ty repr)
    {
	output_decl(Ksi::TyAlias[List::Nil, make_id(name), repr]);
	return Ksi::TyRef[make_id(name)];
    }
    
    void struct2(string name, Ksi::ty t1, string f1, Ksi::ty t2, string f2)
    {
	output_decl(Ksi::StructDef[List::Nil, make_id(name), 
				   List::cons2([t1, make_id(f1)], 
					       [t2, make_id(f2)])]);
    }
    
    void proto0(Ksi::ty ret, string name)
    {
	output_decl(Ksi::FuncProto[List::cons1(Ksi::Extern),
				   ret, make_id(name), List::Nil]);
    }
    
    void proto1(Ksi::ty ret, string name, Ksi::ty arg)
    {
	output_decl(Ksi::FuncProto[List::cons1(Ksi::Extern),
				   ret, make_id(name), List::cons1(arg)]);
    }
    
    void proto2(Ksi::ty ret, string name, Ksi::ty arg1, Ksi::ty arg2)
    {
	output_decl(Ksi::FuncProto[List::cons1(Ksi::Extern),
				   ret, make_id(name), List::cons2(arg1, arg2)]);
    }
    
    struct2("gont_string", 
	    ptroff, "size",
	    Ksi::UInt[8], "first_char");
    
    struct2("gont_lambda_expr", 
	    poly, "func_ptr",
	    poly, "closure");
	    
    struct2("gont_union", 
	    ptroff, "tag",
	    poly, "value");
	    
    gt_int = ini("g_int", flat_type(Int, true));
    gt_void = ini("g_void", flat_type(Void, true));
    gt_exn = ini("g_exn", flat_type(Exn, true));
    gt_string = ini("g_string", flat_type(String, true));
    gt_bool = ini("g_bool", flat_type(Bool, true));
    gt_float = ini("g_float", flat_type(Float, true));

    gt_tbool = ini("g_bool_t", flat_type(Bool, false));
    gt_tfloat = ini("g_float_t", flat_type(Float, false));
    gt_tvoid = ini("g_void_t", flat_type(Void, false));
    
    // gont_str{join,eq,cmp},
    // gont_make_{union,float,string,lambda_expr}
    // gont_match_failure, gont_raise, gont_push_try_block, 
    // gont_pop_raised_value

    proto2(gont_string, "gont_strjoin", gont_string, gont_string);
    proto2(ptroff, "gont_strcmp", gont_string, gont_string);
    proto2(ptroff, "gont_streq", gont_string, gont_string);

    proto2(gont_union, "gont_make_union", ptroff, poly);
    proto1(gt_tfloat, "gont_make_float", gt_float);
    proto2(gont_string, "gont_make_string", ptroff, Ksi::Ptr[Ksi::UInt[8]]);
    proto2(gont_lambda_expr, "gont_make_lambda", poly, poly);
    proto2(poly_ptr, "gont_make_tuple", ptroff, Ksi::Ellipsis);
    
    // TODO: set them no_return
    proto0(ksi_void, "gont_match_failure");
    proto1(ksi_void, "gont_raise", gt_exn);
    proto0(ksi_void, "gont_null_access");
    
    proto1(poly_ptr, "gont_xmalloc", Ksi::MacroType["uptr_off"]);
    
    proto0(Ksi::Ptr[Ksi::Void], "gont_push_try_block");
    proto0(gt_exn, "gont_pop_raised_value");
}

Ksi::ty ksi_type(ty x, bool storage)
{
    Ksi::ty real_type(ty t)
    {
	switch t {
	case NamedType[_]:
	    return ksi_type(type_instance_one(t), storage);
	case _:
	    if (is_type_instance(t))
		return flat_type(t, storage);
	    else
		return real_type(type_instance_one(t));
	}
    }
    
    Ksi::id push_tyalias(ty t)
    {
	string m = mangle_type(t);

	if (storage)
	    m = "S:" + m;
	else
	    m = "T:" + m;
	
	
	if (Hashtbl::mem(mangled_types, m))
	    return Hashtbl::find(mangled_types, m);
	else {
	    Ksi::id i = make_id("t" + itoa(mangled_id++));
	    Hashtbl::add(mangled_types, m, i);
	    output_decl(Ksi::TyAlias[List::cons1(Ksi::ExportName[m]), i,
				     real_type(t)]);
	    return i;
	}
    }
    
    switch x {
    case Int:	    return gt_int;
    case String:    return gt_string;
    case Exn:	    return gt_exn;
    case Bool:
	if (storage)	return gt_bool;
	else		return gt_tbool;
    case Float:	    
	if (storage)	return gt_float;
	else		return gt_tfloat;
    case Void:	    
	if (storage)	return gt_void;
	else		return gt_tvoid;

    case _:	    return Ksi::TyRef[push_tyalias(x)];
    }

}
// Ksi type used to locally store value of type [t]
Ksi::ty storage_type(ty t)
{
    return ksi_type(t, true);
}

// Ksi type used to pass value of type [t] to function
Ksi::ty transport_type(ty t)
{
    return ksi_type(t, false);
}

bool need_t2s_conversion(ty t)
{
    switch type_instance(t) {
    case Bool:	    return true;
    case Float:	    return true;
    case _:	    return false;
    }
}

bool need_s2t_conversion(ty t)
{
    return need_t2s_conversion(t);
}

Ksi::expr call0(string name)
{
    return Ksi::Call[make_id(name), List::Nil];
}

Ksi::expr call1(string name, Ksi::expr e)
{
    return Ksi::Call[make_id(name), List::cons1(e)];
}

Ksi::expr call2(string name, Ksi::expr e1, Ksi::expr e2)
{
    return Ksi::Call[make_id(name), List::cons2(e1, e2)];
}

Ksi::expr gont_alloc(Ksi::expr sz)
{
    return call1("gont_xmalloc", sz);
}

Ksi::expr bin(Ksi::binop o, Ksi::expr e1, Ksi::expr e2)
{
    return Ksi::Binary[o, e1, e2];
}

Ksi::expr s2t_conversion(ty t, Ksi::expr e)
{
    switch type_instance(t) {
    case Bool:
	return Ksi::Convert[transport_type(Bool), e];
	
    case Float:	    
	return call1("gont_make_float", e);
	
    case _:	    
	return e;
    }
}

Ksi::expr t2s_conversion(ty t, Ksi::expr e)
{
    switch type_instance(t) {
    case Bool:
	return Ksi::Convert[storage_type(Bool), e];
	
    case Float:	    
	return Ksi::Indirect[Ksi::Convert[gt_tfloat, e]];
	
    case _:	    
	return e;
    }
}

string parent_closure_type_name(cg_ctx c)
{
    switch c.parent_func {
    case Some[f]:
	return f.cf_closure_type_name;
    case None:
	return "";
    }
}

Ksi::expr pair(Ksi::expr e1, Ksi::expr e2)
{
    return Ksi::Compound[List::cons2(e1, e2)];
}

Ksi::expr bare_vardef_init(Ksi::ty t, Ksi::id n, Ksi::expr ini)
{
    return pair(Ksi::Var[t, n], Ksi::Modify[Ksi::Ref[n], ini]);
}

Ksi::expr local_ref(var_symbol s)
{
    let r = Ksi::Ref[make_id(s.name)] in
    if (has_property(s, IsInClosure))
	return Ksi::Indirect[r];
    else
	return r;
}

<Ksi::attr>List::t public_attrs(bool is_public)
{
    if (is_public)
	return List::cons1(Ksi::Public);
    else
	return List::cons1(Ksi::Local);
}

void push_union(var_symbol u)
{
    int cnt = 1;
    
    var_symbol get_name(<var_symbol>field f)
    {
	return get_binding(f.fld_symbol);
    }

    *[Ksi::id, int] get_enum_entry(var_symbol s)
    {
	set_property(s, IsPrototyped);
	return [make_id(s.a.extern_name), cnt++];
    }
    
    switch (union_of(u)).stype {
    case Union[lst]:
	let fields = List::map(get_name, lst) in
	let content = List::map(get_enum_entry, fields) in
	output_decl(Ksi::EnumDef[content]);
    case _:
	ice("Codegen::push_union(): not union");
    }
}

void push_proto(var_symbol s)
{
    if (has_property(s, IsPrototyped))
	return;

    if (has_property(s, IsException)) {
	let ex = List::cons1(Ksi::Extern) in
	output_decl(Ksi::Global[ex, ptroff, s.a.extern_name]);
    } else if (has_property(s, IsUnionField)) {
	push_union(s);
    } else {
	bool is_xc = has_property(s, IsExternC);
	bool is_public = is_xc || has_property(s, IsExported) || 
			 s.a.fnc_level < 0;

	let attrs = public_attrs(is_public) in
	let name = make_id(s.a.extern_name) in
	if (is_global_function(s)) {
	    Ksi::ty proto_parm(*[ty, <id>option] p)
	    {
		let [t, _] = p in
		let tt = transport_type(t) in {
//		    dbg("tag: " + itoa(Hacks::union_tag(tt)) + "\n");
		    return tt;
		}
	    }

	    let ft = get_ft_record(type_instance(s.stype)) in
	    let rt = transport_type(ft.ret_type) in
	    let parms = List::map(proto_parm, ft.parms) in {
		let decl = Ksi::FuncProto[attrs, rt, name, parms] in
		output_decl(decl);
	    }
	} else {
	    let t = storage_type(s.stype) in
	    if (!is_xc && s.module == Bind::current_module)
		output_decl(Ksi::Global[attrs, t, name]);
	    else
		let ex = List::Cons[Ksi::Extern, attrs] in
		output_decl(Ksi::Global[ex, t, name]);
	}
    }

    set_property(s, IsPrototyped);
}

Ksi::expr make_lambda_expr(Ksi::expr func_ptr, Ksi::expr closure)
{
    return call2("gont_make_lambda", func_ptr, closure);
}

void push_static_lambda(string fnc_name, string lambda_name)
{
    let t = Ksi::Struct[make_id("gont_lambda_expr")] in
    let id = make_id(lambda_name) in
    let gd = Ksi::Global[List::cons1(Ksi::Local), t, id] in 
    let ptrref = Ksi::Component[Ksi::Ref[id], make_id("func_ptr")] in
    let ptr = Ksi::Convert[Ksi::Poly, 
			   Ksi::FuncRef[make_id(fnc_name)]] in
    let ini = Ksi::Modify[ptrref, ptr] in {
	output_decl(gd);
	static_inits = List::Cons[ini, static_inits];
    }
}

Ksi::expr global_ref(var_symbol s)
{
    push_proto(s);

    Ksi::decl closurize(var_symbol ss)
    {
	int cnt = 0;
	
	*[Ksi::ty, Ksi::id] cl_parm(*[ty, <id>option] p)
	{
	    let [t, _] = p in
	    return [transport_type(t), make_id("_" + itoa(cnt++))];
	}

	int cnt2 = 0;
	
	Ksi::expr ref_parm(*[ty, <id>option] p)
	{
	    return Ksi::Ref[make_id("_" + itoa(cnt2++))];
	}

	let ft = get_ft_record(s.stype) in
	let parms = List::Cons[[null_type, dummy_id],
			       List::map(cl_parm, ft.parms)] in
	let rt = transport_type(ft.ret_type) in
	let name = make_id(s.a.extern_name + "$cl") in {
	    Ksi::expr body = Ksi::Call[make_id(s.a.extern_name), 
				       List::map(ref_parm, ft.parms)];
				       
	    switch type_instance(ft.ret_type) {
	    case Void:
		skip;
	    case _:
		body = Ksi::Return[body];
	    }

	    body = Ksi::Bind[body];

	    return Ksi::Func[List::Nil, rt, name, parms, body];
	}
    }

    if (has_property(s, IsUnionField))
	return ice("Codegen::global_ref(): union field requested");

    if (is_global_function(s)) {
	if (!has_property(s, IsClosurized)) {
	    output_decl(closurize(s));
	    set_property(s, IsClosurized);
	    let xn = s.a.extern_name in
	    push_static_lambda(xn + "$cl", xn + "$scl");
	}
	let ref = Ksi::Ref[make_id(s.a.extern_name + "$scl")] in
	return Ksi::Unary[Ksi::Addr, ref];
    } else {
	return Ksi::Ref[make_id(s.a.extern_name)];
    }
}

Ksi::expr raddr(Ksi::expr e)
{
    return Ksi::Unary[Ksi::RAddr, e];
}

Ksi::expr compile_ref(cg_ctx c, ref_record rr)
{
    let s = get_binding(rr.sym) in
    if (s.a.fnc_level <= 0) {
	return global_ref(s);
    } else if (rr.level == 0) {
	return local_ref(s);
    } else {
	if (!Ptrhash::mem(c.cf_references_hash, s)) {
	    Ksi::expr go_up(int n)
	    {
		if (n == 1)
		    return parent_closure_ref;
		else
		    return Ksi::Arrow[go_up(n - 1), parent_closure_id];
	    }
	  
	    let in_parent = Ksi::Arrow[go_up(rr.level),
				       make_id(s.a.extern_name)] in
	    let vi = bare_vardef_init(Ksi::Reference[storage_type(s.stype)], 
				      make_id(s.name), 
				      raddr(in_parent)) in {
		Ptrhash::add(c.cf_references_hash, s, 1);
		c.cf_references = List::Cons[vi, c.cf_references];
	    }
	}

	return Ksi::Indirect[Ksi::Ref[make_id(s.name)]];
    }
}

Ksi::expr vardef(cg_ctx c, var_symbol s)
{
    let t = storage_type(s.stype) in
    let n = make_id(s.name) in
    if (has_property(s, IsInClosure)) {
	let in_closure = Ksi::Arrow[our_closure_ref, 
				    make_id(s.a.extern_name)] in
	return bare_vardef_init(Ksi::Reference[t], n, raddr(in_closure));
    } else {
	return Ksi::Var[t, n];
    }
}

Ksi::expr vardef_init(cg_ctx c, var_symbol s, Ksi::expr ini)
{
    return pair(vardef(c, s), Ksi::Modify[local_ref(s), ini]);
}

void push_closure_type(cg_ctx c, <var_symbol>List::t closure_vars)
{
    int cnt = 1;
    
    *[Ksi::ty, Ksi::id] compile_cv(var_symbol s)
    {
	s.a.extern_name = s.name + "$" + itoa(cnt++);
	return [storage_type(s.stype), make_id(s.a.extern_name)];
    }
   
//dbg("fnc: " + c.cf_name + ": " + itoa(List::length(closure_vars)) + "\n");

    <*[Ksi::ty, Ksi::id]>List::t vars = 
		 List::map(compile_cv, closure_vars);

    let p = parent_closure_type_name(c) in
    if (p != "") {
	let fld = [Ksi::Ptr[Ksi::Struct[make_id(p)]], parent_closure_id] in
	vars = List::Cons[fld, vars];
    }
	
    if (List::length(vars) == 0) {
	ice("Codegen::push_closure_type(): needs closure but has none");
    } else {
	c.cf_closure_type_name = c.cf_name + "$closure_struct";
	let id = make_id(c.cf_closure_type_name) in
	output_decl(Ksi::StructDef[List::Nil, id, vars]);
    }
}


Ksi::expr wfl_locate(cg_ctx c, *(Ksi::expr (cg_ctx, 'a)) f, <'a>located v)
{
    push_location(v.loc);
    try {
	let r = f(c, v.a) in
	return Ksi::Line[v.loc.first_line, r];
    } finally {
	pop_location();
    }
}

Ksi::binop ksi_op2(op2 o)
{
    switch o {
    case Plus:	    return Ksi::Plus;
    case Minus:	    return Ksi::Minus;
    case Mult:	    return Ksi::Mult;
    case Mod:	    return Ksi::TruncMod;
    case Div:	    return Ksi::TruncDiv;
    case Eq:	    return Ksi::Eq;
    case Ne:	    return Ksi::Ne;
    case Lt:	    return Ksi::Lt;
    case Gt:	    return Ksi::Gt;
    case Ge:	    return Ksi::Ge;
    case Le:	    return Ksi::Le;
    case AndIf:	    return Ksi::TruthAndIf;
    case OrIf:	    return Ksi::TruthOrIf;
    case Assign:    return ice("Codegen::ksi_op2(): assign");
    }
}

// conservative estimation, allocation is not counted as side-effect
bool has_side_effects(expr ex)
{
    switch ex.a.raw {
    case SimpleBinOp[Assign, _, _]:
	return true;
	
    case SimpleBinOp[_, e1, e2]:
	return has_side_effects(e1) || has_side_effects(e2);
	
    case SimpleUnOp[op, e]:
	switch op {
	case Negate:	return has_side_effects(e);
	case PreDec:	return true;
	case PreInc:	return true;
	case PostDec:	return true;
	case PostInc:	return true;
	}
    
    case Ref[_]:
	return false;

    case Call[_, _]:
	return true;

    case TupleExpr[lst]:
	return List::exists(has_side_effects, lst);

    case UnionExpr[u]:
	return has_side_effects(u.nx_expr);
    
    // we lay a bit here, as generating lambda is lots of work
    // in case of lack of closure
    case Lambda[_]:
	return true;

    case StructExpr[lst]:
	return List::exists(fun bool (sini_expr s) 
				(has_side_effects(s.sx_expr)),
			    lst);

    case DotExpr[de]:
	return has_side_effects(de.sx_expr);
    
    case BoolConst[_]:	    return false;
    case IntConst[_]:	    return false;
    case FloatConst[x]:	    return false;
    case StringConst[x]:    return false;
    case VoidConst:	    return false;
    case NullConst:	    return false;
    }
}

bool is_poly(ty tt)
{
    bool is = false;
    
    void poly_check(ty t)
    {
	switch t {
	case TyVar[_]:
	    is = true;
	case XTyVar[_]:
	    is = true;
	case _:
	    skip;
	}
    }

    type_iter(poly_check, tt);

    return is;
}

// opt-TODO: we can skip saving of last expression with side-effects
*[<Ksi::expr>List::t, <Ksi::expr>option] call_args(cg_ctx c, <expr>List::t args,
						   <ty>List::t formal_types)
{
    Ksi::expr ini = nop;
    int pos = 0;
    
    Ksi::expr map_arg(expr e)
    {
	Ksi::expr x = compile_expr(c, e);
	
	if (has_side_effects(e)) {
	    let t = storage_type(type_of(e)) in
	    let n = make_id("arg" + itoa(pos++) + "$") in {
		ini = pair(ini, bare_vardef_init(t, n, x));
		x = Ksi::Ref[n];
	    }
	}

	x = s2t_conversion(type_of(e), x);

	switch formal_types {
	case List::Cons[hd, tl]:
	    formal_types = tl;
	    if (is_poly(hd))
		x = Ksi::Convert[transport_type(hd), x];
		
	case List::Nil:
	    skip;
	}

	return x;
    }
   
    let ret = List::map(map_arg, args) in
    switch ini {
    case Ksi::Nop:
	return [ret, None];
    case _:
	return [ret, Some[ini]];
    }
}

Ksi::expr maybe_save(cg_ctx c, expr to_save, *(Ksi::expr (Ksi::expr)) in_what)
{
    let compiled = compile_expr(c, to_save) in
    switch to_save.a.raw {
    case Ref[_]:
	// ok, we don't bother saving
	return in_what(compiled);
	
    case _:
	// but here we should, or even have to (if expression has
	// side effects)
	let n = make_id("saved$") in
	let t = storage_type(type_of(to_save)) in
	let save = bare_vardef_init(t, n, compiled) in
	return Ksi::Bind[pair(save, in_what(Ksi::Ref[n]))];
    }
}

Ksi::expr compile_call(cg_ctx c, ty our_type, expr fnc, <expr>List::t args)
{
    <Ksi::expr>option ini;
    <Ksi::expr>List::t cargs;

    let formal_types = List::map(fst, (get_ft_record(type_of_s(fnc))).parms) in
    [cargs, ini] = call_args(c, args, formal_types);
    
    Ksi::expr lambda_call()
    {
	let ft = get_ft_record(type_of_s(fnc)) in
	let targs = List::map(transport_type, List::map(fst, ft.parms)) in
	let cl_args = List::Cons[Ksi::Poly, targs] in {
	    Ksi::ty fnc_type = Ksi::FuncPtr[transport_type(ft.ret_type), 
					    cl_args];

	    Ksi::expr make_call(Ksi::expr ptr)
	    {
		let fnc_ptr = Ksi::Arrow[ptr, make_id("func_ptr")] in
		let closure = Ksi::Arrow[ptr, make_id("closure")] in
		let allargs = List::Cons[closure, cargs] in
		return Ksi::PtrCall[Ksi::Convert[fnc_type, fnc_ptr], allargs];
	    }

	    return maybe_save(c, fnc, make_call);
	}
    }
    
    Ksi::expr call;
    
    // optimize calls to global functions
    switch fnc.a.raw {
    case Ref[rr]:
	let s = get_binding(rr.sym) in
	if (is_global_function(s)) {
	    push_proto(s);
	    call = Ksi::Call[make_id(s.a.extern_name), cargs];
	} else {
	    call = lambda_call();
	}
    case _:
	call = lambda_call();
    }

    call = t2s_conversion(our_type, call);

    switch ini {
    case Some[i]:
	call = Ksi::Bind[pair(i, call)];
    case None:
	skip;
    }

    if (is_poly(type_of(fnc)))
	call = Ksi::Convert[storage_type(our_type), call];

    return call;
}

Ksi::expr compile_tuple(cg_ctx c, <expr>List::t args)
{
    let [a, i] = call_args(c, args, List::Nil) in
    let cnt = Ksi::IntConst[ptroff, Int64::of_int(List::length(args))] in
    let call = Ksi::Call[make_id("gont_make_tuple"), List::Cons[cnt, a]] in {
	switch i {
	case Some[ini]:
	    return Ksi::Bind[pair(ini, call)];
	    
	case None:
	    return call;
	}
    }
}

Ksi::expr compile_struct_expr(cg_ctx c, ty our_type, <sini_expr>List::t inis)
{
    Ksi::expr sz;

    switch flat_type(our_type, true) {
    case Ksi::Ptr[Ksi::Struct[s]]:
	sz = Ksi::SizeOf[Ksi::Struct[s]];
    case _:
	return ice("Codegen::compile_struct_expr(): flat_type went crazy");
    }
    
    let n = make_id("sx$") in
    let alloc = Ksi::Convert[storage_type(our_type), gont_alloc(sz)] in
    let v = bare_vardef_init(storage_type(our_type), n, alloc) in {
	Ksi::expr ref = Ksi::Ref[n];
	Ksi::expr ini = v;
	
	void init_one(sini_expr se)
	{
	    let s = get_binding(se.sx_symbol) in
	    let x = compile_expr(c, se.sx_expr) in
	    let i = Ksi::Modify[Ksi::Arrow[ref, make_id(s.name)], x] in
	    ini = pair(ini, i);
	}
	List::iter(init_one, inis);
	return Ksi::Bind[pair(ini, ref)];
    }
}

Ksi::expr symbol_union_tag(var_symbol s)
{
    let id = make_id(s.a.extern_name) in
    if (has_property(s, IsException))
	return Ksi::Ref[id];
    else
	return Ksi::EnumRef[id];
}

Ksi::expr compile_union_expr(cg_ctx c, namedexpr u)
{
    let s = get_binding(u.nx_symbol) in {
	push_proto(s);
	
	Ksi::expr tag = symbol_union_tag(s);
	
	void push_static()
	{
	    let id = make_id(s.a.extern_name + "$sa") in
	    let decl = Ksi::Global[List::cons1(Ksi::Local),
				   Ksi::Struct[make_id("gont_union")],
				   id] in
	    let ini = Ksi::Modify[Ksi::Component[Ksi::Ref[id], union_tag_id],
				  tag] in 
	    {
		output_decl(decl);
		static_inits = List::Cons[ini, static_inits];
		set_property(s, IsStaticallyAllocated);
	    }
	}
	
	switch u.nx_expr.a.raw {
	case VoidConst:  
	    if (!has_property(s, IsStaticallyAllocated))
		push_static();

	    return Ksi::Unary[Ksi::Addr, Ksi::Ref[s.a.extern_name + "$sa"]];
	case _:
	    let t = type_of(u.nx_expr) in
	    let ex = s2t_conversion(t, compile_expr(c, u.nx_expr)) in
	    return call2("gont_make_union", tag, ex);
	}

    }
}

Ksi::expr tuple_nth(Ksi::expr e, int n)
{
    let pos = bin(Ksi::Mult, 
		  Ksi::IntConst[ptroff, Int64::of_int(n)],
		  Ksi::Convert[ptroff, Ksi::SizeOf[ptroff]]) in
    return Ksi::Indirect[bin(Ksi::Plus, Ksi::Convert[poly_ptr, e], pos)];
}

Ksi::expr union_tag(Ksi::expr e)
{
    return Ksi::Arrow[Ksi::Convert[gont_union, e], union_tag_id];
}

Ksi::expr union_value(Ksi::expr e)
{
    return Ksi::Arrow[Ksi::Convert[gont_union, e], union_value_id];
}

Ksi::expr and(Ksi::expr e1, Ksi::expr e2)
{
    return bin(Ksi::TruthAndIf, e1, e2);
}

Ksi::expr eq(Ksi::expr e1, Ksi::expr e2)
{
    return bin(Ksi::Eq, e1, e2);
}

Ksi::expr compile_assign(cg_ctx c, expr dst, expr src)
{
    switch dst.a.raw {
    case TupleExpr[_]:
	let n = make_id("tuple_assign$") in
	let v = bare_vardef_init(poly_ptr, n, compile_expr(c, src)) in
	let r = Ksi::Ref[n] in {
	    Ksi::expr assign = v;
	    
	    void walk(Ksi::expr handle, expr d)
	    {
		switch d.a.raw {
		case TupleExpr[lst]:
		    int pos = 0;
		    List::iter(fun void (expr x) 
				    {walk(tuple_nth(handle, pos++), x);},
				lst);
		case _:
		    assign = pair(assign,
				  Ksi::Modify[compile_expr(c, d), handle]);
		}
	    }

	    walk(r, dst);

	    return Ksi::Bind[assign];
	}

    case _:
	return Ksi::Modify[compile_expr(c, dst), compile_expr(c, src)];
    }
}

Ksi::expr compile_dot_expr(cg_ctx c, expr ex, sf_symbol field)
{
    Ksi::id field_name = make_id(field.name);

    switch type_of_s(ex) {
    case OptStruct[_]:
	Ksi::expr make_check(Ksi::expr ptr)
	{
	    let t = storage_type(type_of(ex)) in
	    let check = Ksi::Cond[eq(ptr, null_value), 
				  call0("gont_null_access"),
				  nop] in
	    return pair(check, Ksi::Arrow[ptr, field_name]);
	}
	
	return maybe_save(c, ex, make_check);
    
    case Struct[_]:
	return Ksi::Arrow[compile_expr(c, ex), field_name];

    case _:
	return ice("Codegen::compile_dot_expr(): non struct type");
    }
}

Ksi::expr compile_nl_expr(cg_ctx c, nl_expr nle)
{
    switch nle.raw {
    case SimpleBinOp[Assign, e1, e2]:
	return compile_assign(c, e1, e2);
    
    case SimpleBinOp[op, e1, e2]:
	let x1 = compile_expr(c, e1) in
	let x2 = compile_expr(c, e2) in
	switch [type_of_s(e1), op] {
	case [String, Plus]:
	    return call2("gont_strjoin", x1, x2);
	    
	case [String, Eq]:
	    return Ksi::Convert[Ksi::Bool, call2("gont_streq", x1, x2)];
	    
	case [String, Ne]:
	    return eq(call2("gont_streq", x1, x2), zero_const);
	
	case [String, _]:
	    return bin(ksi_op2(op), call2("gont_strcmp", x1, x2), zero_const);
	
	case [Float, Div]:
	    return bin(Ksi::RDiv, x1, x2);
	
	case [_, _]:
	    return bin(ksi_op2(op), x1, x2);
	}
	
    case SimpleUnOp[op, e]:
	let x = compile_expr(c, e) in
	switch op {
	case Negate:	return Ksi::Unary[Ksi::Negate, x];
	case PreDec:	return Ksi::Binary[Ksi::PreDec, x, one_const];
	case PreInc:	return Ksi::Binary[Ksi::PreInc, x, one_const];
	case PostDec:	return Ksi::Binary[Ksi::PostDec, x, one_const];
	case PostInc:	return Ksi::Binary[Ksi::PostInc, x, one_const];
	}
    
    case Ref[rr]:
	return compile_ref(c, rr);

    case Call[e, el]:
	let our_type = type_of(({loc = top_location(), a = nle})) in
	return compile_call(c, our_type, e, el);

    case TupleExpr[el]:
	return compile_tuple(c, el);

    case UnionExpr[u]:
	return compile_union_expr(c, u);
    
    case Lambda[l]:
	return compile_lambda(c, l);

    case StructExpr[s]:
	let our_type = type_of(({loc = top_location(), a = nle})) in
	return compile_struct_expr(c, our_type, s);

    case DotExpr[de]:
	return compile_dot_expr(c, de.sx_expr, get_binding(de.sx_symbol));
    
    case BoolConst[x]:
	return Ksi::BoolConst[x];
    
    case IntConst[x]:
	return Ksi::IntConst[gt_int, x];
    
    case FloatConst[x]:
	return Ksi::FloatConst[gt_float, x];
	
    case StringConst[x]:
	let len = Ksi::IntConst[ptroff, Int64::of_int(String::length(x))] in
	return call2("gont_make_string", len, Ksi::StringConst[x]);
	
    case VoidConst:
	return ice("Codegen::compile_nl_expr(): void-const");
	
    case NullConst:
	return null_value;
    }
}

Ksi::expr compile_expr(cg_ctx c, expr e)
{
    return wfl_locate(c, compile_nl_expr, e);
}

*[<Ksi::expr>option, Ksi::expr] 
match_and_init(cg_ctx c, Ksi::expr handle, pattern p, bool transport)
{
    switch p.a {
    case VoidPattern:
	return [None, nop];
	
    case VarPattern[v]:
	Ksi::expr val;
	let s = get_binding(v.var_name) in {
	    if (transport) {
		let t = s.stype in
		val = Ksi::Convert[storage_type(t), 
				   t2s_conversion(t, handle)];
	    } else {
		val = handle;
	    }

	    if (has_property(s, IsGuardLocal) || 
		has_property(s, IsLetVariable))
		return [None, vardef_init(c, s, val)];
	    else
		return [None, Ksi::Modify[local_ref(s), val]]; 
      }
	
    case ExprPattern[e]:
	Ksi::expr ex = compile_expr(c, e);
	if (transport)
	    let t = type_of_s(e) in
	    handle = Ksi::Convert[storage_type(t), 
				  t2s_conversion(t, handle)];
	switch type_of_s(e) {
	case String:
	    return [Some[Ksi::Convert[Ksi::Bool, 
				      call2("gont_streq", ex, handle)]], nop];
	case _:
	    return [Some[eq(ex, handle)], nop];
	}
	
    case TuplePattern[lst]:
	Ksi::expr match = nop;
	Ksi::expr ini = nop;
	int pos = 0;
	
	void match_one(pattern pp)
	{
	    switch match_and_init(c, tuple_nth(handle, pos), pp, true) {
	    case [None, ex]:
		ini = pair(ini, ex);

	    case [Some[x], ex]:
		ini = pair(ini, ex);
		switch match {
		case Ksi::Nop:	match = x;
		case _:		match = and(match, x);
		}
	    }

	    pos++;
	}

	List::iter(match_one, lst);
	
	switch match {
	case Ksi::Nop:	return [None, ini];
	case _:		return [Some[match], ini];
	}

    case UnionPattern[u]:
	let s = get_binding(u.np_symbol) in
	let tag = symbol_union_tag(s) in
	let cond = eq(union_tag(handle), tag) in {
	    push_proto(s);
	    switch match_and_init(c, union_value(handle), u.np_pattern, true) {
	    case [None, ini]:
		return [Some[cond], ini];
	    case [Some[ex], ini]:
		return [Some[and(cond, ex)], ini];
	    }
	}
    }
}

Ksi::expr compile_matching(cg_ctx c, Ksi::expr handle, <case_clause>List::t clauses,
			   Ksi::expr header, Ksi::expr all_failed)
{
    Ksi::id block_id = make_id(genname("match_block"));
    Ksi::expr trailer = Ksi::ExitBlock[block_id];
    
    *[Ksi::expr, /* cannot_fail */ bool] compile_case_clause(case_clause cc)
    {
	bool cannot_fail = false;
	Ksi::expr test_alternatives = nop;

	void compile_one_case(*[pattern, expr] oc)
	{
	    let [pat, cond] = oc in {
		if (cannot_fail) {
		    if (flags.wpattern)
			warnl(pat.loc, "matching code compiled out");
		    return;
		}
		
		Ksi::expr test;
		
		let [match, ini] = match_and_init(c, handle, pat, false) in {
		    switch [match, cond.a.raw] {
		    case [None, BoolConst[true]]:
			cannot_fail = true;
			test = Ksi::Bind[pair(ini, true_const)];
			
		    case [Some[x], _]:
			let ccond = compile_expr(c, cond) in
			test = Ksi::Bind[and(x, pair(ini, ccond))];
			
		    case [None, _]:
			let ccond = compile_expr(c, cond) in
			test = Ksi::Bind[pair(ini, ccond)];
		    }
		}

		switch test_alternatives {
		case Ksi::Nop:
		    test_alternatives = test;
		    
		case x:
		    test_alternatives = Ksi::Binary[Ksi::TruthOrIf, x, test];
		}
	    }
	}

	List::iter(compile_one_case, cc.cases);

	let body = compile_stmt(c, cc.body) in
	let all = Ksi::Compound[List::cons3(header, body, trailer)] in
	let vardefs = Ksi::Compound[List::map(partial(vardef, c), cc.patvars)] in
	let cond = Ksi::Cond[test_alternatives, all, nop] in
	return [Ksi::Bind[pair(vardefs, cond)], cannot_fail];
    }

    <Ksi::expr>List::t compile_clauses(<case_clause>List::t cl)
    {
	switch cl {
	case List::Cons[cc, tl]:
	    let [ex, cannot_fail] = compile_case_clause(cc) in
	    if (cannot_fail) {
		if (flags.wpattern) {
		    switch tl {
		    case List::Cons[hd, _]:
			warnl(hd.body.loc, "matching code compiled out");
		    case List::Nil:
			skip;
		    }
		}
		
		return List::cons1(ex);
	    } else {
		return List::Cons[ex, compile_clauses(tl)];
	    }
	    
	case List::Nil:
	    return List::cons1(all_failed);
	}
    }

    return Ksi::Block[block_id, Ksi::Compound[compile_clauses(clauses)]];
}

Ksi::expr voidify(Ksi::expr e)
{
    return Ksi::Convert[Ksi::Void, e];
}

Ksi::expr compile_switch(cg_ctx c, expr e, <case_clause>List::t clauses)
{
    let sw_id = make_id("sw$") in
    let vi = bare_vardef_init(storage_type(type_of(e)), sw_id, 
			      compile_expr(c, e)) in
    let fail = call0("gont_match_failure") in
    let matching = compile_matching(c, Ksi::Ref[sw_id], clauses, nop, fail) in
    return Ksi::Bind[pair(vi, matching)];
}

Ksi::expr compile_let(cg_ctx c, <*[pattern, expr]>List::t assigns, stmt s)
{
    Ksi::expr conds = nop;
    Ksi::expr inis = nop;
    int cnt = 0;

    void one_assign(*[pattern, expr] assign)
    {
	let [pat, ex] = assign in
	let n = make_id("let$" + itoa(cnt++)) in
	let v = bare_vardef_init(storage_type(type_of(ex)), 
				 n, compile_expr(c, ex)) in
	switch match_and_init(c, Ksi::Ref[n], pat, false) {
	case [Some[cond], ini]:
	    let maybe_fail = Ksi::Cond[cond, nop, 
				       call0("gont_match_failure")] in
	    conds = pair(conds, pair(v, maybe_fail));
	    inis = pair(inis, ini);
	    
	case [None, ini]:
	    conds = pair(conds, v);
	    inis = pair(inis, ini);
	}
    }

    List::iter(one_assign, assigns);

    return Ksi::Bind[pair(pair(conds, inis), compile_stmt(c, s))];
}

Ksi::expr aint_null(Ksi::expr e)
{
    return Ksi::Binary[Ksi::Ne, e, null_value];
}

/*
    exn = NULL
    try
      if push_ctx == local
        try
          if push_ctx == local
            TRY_BODY
        finally
          exn = pop_ctx
	  if exn
            WITH_BODY
          if exn
            raise exn
    finally
      exn = pop_ctx
      FINALLY_BODY
      if exn
        raise exn
*/

Ksi::expr compile_try(cg_ctx c, try_record tr)
{
    Ksi::id exn_value_id = make_id("exn$");
    Ksi::expr exn_ref_expr = Ksi::Ref[exn_value_id];

    Ksi::expr try_finally(Ksi::expr body, Ksi::expr fin)
    {
	Ksi::expr if_local_return(Ksi::expr e)
	{
	    let args = List::cons1(call0("gont_push_try_block")) in
	    let cond = eq(Ksi::BuiltIn[make_id("setjmp"), args], zero_const) in
	    return Ksi::Cond[cond, e, nop];
	}
	
	let pop = Ksi::Modify[exn_ref_expr,
			      call0("gont_pop_raised_value")] in
	let maybe_reraise = Ksi::Cond[aint_null(exn_ref_expr),
				      call1("gont_raise", exn_ref_expr),
				      nop] in
	return Ksi::TryFinally[if_local_return(body), 
			       pair(pair(pop, fin), maybe_reraise)];
    }

    Ksi::expr with_matching(<case_clause>List::t clauses)
    {
	let kill_exn = Ksi::Modify[exn_ref_expr, null_value] in
	let name = make_id("exn_copy$") in
	let vi = bare_vardef_init(storage_type(Exn), name, exn_ref_expr) in
	let m = compile_matching(c, Ksi::Ref[name], clauses, kill_exn, nop) in
	return Ksi::Cond[aint_null(exn_ref_expr), pair(vi, m), nop];
    }
    
    let vi = bare_vardef_init(storage_type(Exn), exn_value_id, null_value) in {
	Ksi::expr body = compile_stmt(c, tr.try_body);
	if (List::length(tr.try_with) != 0) {
	    body = try_finally(body, with_matching(tr.try_with));
	}

	switch tr.try_finally.a {
	case Skip:
	    skip;
	case _:
	    body = try_finally(body, compile_stmt(c, tr.try_finally));
	}

	return Ksi::Bind[pair(vi, body)];
    }
}

Ksi::expr compile_nl_stmt(cg_ctx c, nl_stmt raw)
{
    switch raw {
    case Skip:
	return nop;
    
    case Expr[e]:
	return voidify(compile_expr(c, e));
    
    case Return[e]:
	switch type_of_s(e) {
	case Void:
	    switch e.a.raw {
	    case VoidConst:
		return Ksi::Return[nop];
	    case _:
		return pair(compile_expr(c, e), Ksi::Return[nop]);
	    }
	case _:
	    return Ksi::Return[s2t_conversion(type_of(e), compile_expr(c, e))];
	}

    case VarDef[v]:
	return vardef(c, get_binding(v.var_name));
	    
    case Block[lst]:
	let body = List::map(partial(compile_stmt, c), lst) in
	return Ksi::Bind[Ksi::Compound[body]];

    case Pair[s1, s2]:
	return pair(compile_stmt(c, s1), compile_stmt(c, s2));
    
    case Switch[e, clauses]:
	return compile_switch(c, e, clauses);
	    
    case Let[assigns, s]:
	return compile_let(c, assigns, s);
    
    case If[e, s1, s2]:
	return Ksi::Cond[compile_expr(c, e),
			 compile_stmt(c, s1),
			 compile_stmt(c, s2)];
			 
    case Loop[l]:
	let brk = make_id("break$" + simple_name(l.ll_sym)) in
	let cont = make_id("continue$" + simple_name(l.ll_sym)) in
	let body = compile_stmt(c, l.ll_body) in
	let inc = compile_stmt(c, l.ll_inc) in
	return Ksi::Block[brk, Ksi::Loop[pair(Ksi::Block[cont, body], inc)]];
    
    case Break[b]:
	string prefix;
	switch b.kind {
	case BreakKind:	    prefix = "break$";
	case ContinueKind:  prefix = "continue$";
	}
	return Ksi::ExitBlock[make_id(prefix + simple_name(b.sym))];
	    
    case BreakIfNot[e]:
	return Ksi::ExitLoop[Ksi::Unary[Ksi::TruthNot, compile_expr(c, e)]];
    
    case Open[_]:
	return nop;
    
    case Try[tr]:
	return compile_try(c, tr);
    
    case Raise[e]:
	return call1("gont_raise", compile_expr(c, e));
    }
}

Ksi::expr compile_stmt(cg_ctx c, stmt s)
{
    return wfl_locate(c, compile_nl_stmt, s);
}

struct fn_desc {
    string name;
    bool is_public;
    bool takes_closure_parm;
    lambda_record lambda;
}

void push_fn_decl(cg_ctx c, fn_desc fd)
{
    <*[Ksi::ty, Ksi::id]>List::t ksi_parms()
    {
	<var_symbol>List::t parm_syms = fd.lambda.lambda_parms;
	int id = 1;
	
	*[Ksi::ty, Ksi::id] compile_parm(*[ty, <id>option] parm)
	{
	    switch parm {
	    case [_, Some[_]]:
		var_symbol s;
		switch parm_syms {
		case List::Cons[hd, tl]:
		    parm_syms = tl;
		    s = hd;
		case _:
		    return ice("Codegen::push_fn_decl(): not enough symbols");
		}
		
		let t = transport_type(s.stype) in
		if (has_property(s, IsInClosure) || 
		    need_t2s_conversion(s.stype))
		    return [t, make_id(s.name + "$tmp")];
		else
		    return [t, make_id(s.name)];

	    case [t, None]:
		return [transport_type(t), make_id("unused$" + itoa(id++))];
	    }
	    
	}
	
	*[Ksi::ty, Ksi::id] closure_parm()
	{
	    let p = parent_closure_type_name(c) in
	    if (p != "")
		return [Ksi::Ptr[Ksi::Struct[make_id(p)]], parent_closure_id];
	    else
		return [null_type, dummy_id];
	}
	
	let ft = get_ft_record(fd.lambda.lambda_type) in
	let parms = List::map(compile_parm, ft.parms) in
	if (fd.takes_closure_parm)
	    return List::Cons[closure_parm(), parms];
	else
	    return parms;
    }

    Ksi::expr convert_parm(var_symbol s)
    {
	if (has_property(s, IsInClosure)) {
	    let tmp = Ksi::Ref[make_id(s.name + "$tmp")] in
	    return vardef_init(c, s, t2s_conversion(s.stype, tmp));
	} else {
	    if (need_t2s_conversion(s.stype)) {
		let tmp = Ksi::Ref[make_id(s.name + "$tmp")] in
		return vardef_init(c, s, t2s_conversion(s.stype, tmp));
	    } else {
		return nop;
	    }
	}
    }

    // definition of closure variable, allocation of memory for closure
    // and possibly assignment of parent_closure$ field.
    Ksi::expr closure_stuff()
    {
	if (!fd.lambda.lambda_needs_closure) {
	    // ha! no closure
	    c.cf_closure_type_name = "";
	    return nop;
	}
	
	push_closure_type(c, fd.lambda.lambda_closure_vars);
	
	Ksi::expr ret;
	
	let closure_type = Ksi::Struct[make_id(c.cf_closure_type_name)] in
	let clptr = Ksi::Ptr[closure_type] in
	let mem = Ksi::Convert[clptr, gont_alloc(Ksi::SizeOf[closure_type])] in
	ret = bare_vardef_init(clptr, our_closure_id, mem);

	if (parent_closure_type_name(c) != "") {
	    let par = Ksi::Arrow[our_closure_ref, parent_closure_id] in
	    let ini = Ksi::Modify[par, parent_closure_ref] in
	    ret = pair(ret, ini);
	}
	    
	return ret;
    }
    
    c = ({cf_is_public = fd.is_public,
	  cf_name = fd.name,
	  cf_ret_type_instanced = 
		type_instance((get_ft_record(fd.lambda.lambda_type)).ret_type),
	  parent = Some[c],
	  parent_func = Some[c],
	  cf_closure_type_name = "?",
	  cf_references_hash = Ptrhash::create(13),
	  cf_references = List::Nil
    });
    
    
    if (fd.is_public)
        top_comment("public " + fd.name);
    else
	top_comment("local " + fd.name);
    
    let ksi_ret_type = transport_type(c.cf_ret_type_instanced) in
   
    let closure = closure_stuff() in
    let parms_init = List::map(convert_parm, fd.lambda.lambda_parms) in
    let startup = Ksi::Compound[List::Cons[closure, parms_init]] in {
	Ksi::expr body = compile_stmt(c, fd.lambda.lambda_body);
	
	let ref_lst = List::Cons[startup, c.cf_references] in
	body = Ksi::Bind[pair(Ksi::Compound[ref_lst], Ksi::Bind[body])];
	
	Ksi::decl d = Ksi::Func[public_attrs(fd.is_public), 
				ksi_ret_type, 
				make_id(fd.name),
				ksi_parms(), 
				body];
				
	top_comment("body " + fd.name);
	output_decl(d);
	top_comment("end " + fd.name);
    }
}

Ksi::expr compile_lambda(cg_ctx c, lambda_record l)
{
    string name = genname(c.cf_name + "$l");
    
    push_fn_decl(c, ({name = name,
		      is_public = false,
		      takes_closure_parm = true,
		      lambda = l}));
   
    if (c.cf_closure_type_name == "") {
	string sname = genname(c.cf_name + "$scl");
	push_static_lambda(name, sname);
	return Ksi::Unary[Ksi::Addr, Ksi::Ref[make_id(sname)]];
    } else
	return make_lambda_expr(Ksi::FuncRef[make_id(name)], our_closure_ref);
}

void codegen_for_nl_decl(nl_decl d)
{
    cg_ctx c = get_global_ctx();
    
    switch d {
    case FuncDecl[fn]:
	let sym = get_binding(fn.fn_sym) in {
	    push_fn_decl(c, ({name = sym.a.extern_name, 
			      is_public = has_property(sym, IsExported),
			      takes_closure_parm = false,
			      lambda = fn.l}));
	    // don't output new prototype for this function
	    set_property(sym, IsPrototyped);
	}
	
    case SectionDecl[sec_type, l]:
	let nam = genname(sec_type) in {
	    push_fn_decl(c, ({name = nam,
			      is_public = false, 
			      takes_closure_parm = false,
			      lambda = l}));

	    switch sec_type {
	    case "init":
		module_init = List::Cons[nam, module_init];
	    case "fini":
		module_fini = List::Cons[nam, module_fini];
	    case _:
		ice("Codegen::codegen_for_nl_decl(): " +
		    "bad section");
	    }
	}
    
    case GlobalDecl[v]:
	// push proto actually outputs definition instead of declaration
	// for globals from current module
	push_proto(get_binding(v.var_name));
		    
    // symbols are prototyped when used
    case ExternC[_]:	    skip;
    case FuncProto[_]:	    skip;
    case XCFuncProto[_, _]: skip;
    case TyDef[_]:	    skip;

    // code for exceptions is generated during link
    case ExceptionDecl[_]:  skip;

    // these don't generate any code
    case TopLevelOpen[_]:   skip;
    case DummyDecl:	    skip;
    }
}

void codegen_for_decl(decl d)
{
    locate(codegen_for_nl_decl, d);
}

string push_function(<Ksi::expr>List::t exprs, string kind)
{
    void output(string name)
    {
	let body = Ksi::Bind[Ksi::Compound[exprs]] in
	let decl = Ksi::Func[List::cons1(Ksi::Public), Ksi::Void, make_id(name), 
			     List::Nil, body] in
	output_decl(decl);
    }

    switch exprs {
    case List::Nil:
	return "";
    case _:
	string name = Bind::current_module + "$$" + kind + "$";
	output(name);
	return name;
    }
}

*[string, string, string] fini()
{
    <Ksi::expr>List::t call_lst(<Ksi::id>List::t names)
    {
	Ksi::expr call(Ksi::id fnc)
	{
	    return Ksi::Call[fnc, List::Nil];
	}
	
	return List::map(call, List::rev(names));
    }
	
    let ctors = push_function(static_inits, "ctor") in
    let inits = push_function(call_lst(module_init), "init") in
    let finis = push_function(call_lst(module_fini), "fini") in
    let xgo_linked_id = make_id(Bind::current_module + "$$xgo_linked$") in
    let xgo_linked = Ksi::Global[List::cons1(Ksi::Extern), null_type,
				 xgo_linked_id] in
    let reference_xgo = Ksi::Modify[Ksi::Ref[xgo_linked_id], null_value] in
    let refxgo_fn = Ksi::Func[List::cons1(Ksi::Public), Ksi::Void,
			      make_id(Bind::current_module + "$$reference_xgo$"),
			      List::Nil, Ksi::Bind[reference_xgo]] in
    {
	output_decl(xgo_linked);
	output_decl(refxgo_fn);

	Ksi::output_trailer(ksi_sink);

	return [ctors, inits, finis];
    }
}

// vim:sts=4:sw=4
