/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 539 $ $LastChangedDate: 2002-08-25 13:25:13 +0200 (Sun, 25 Aug 2002) $
 *
 * Util - Various utility functions: diagnostics, compilation flags, etc.
 *
 * Maybe we should just leave types in Ast and move functions here?
 */
 
open Ast;
open Util;

/*** Flags. ***/
compiler_flags flags;

section init 
{
	flags = ({
		werror = false,
		wshadow = true,
		wdead = true,
		wlabeled = true,
		wpattern = true,
		wcapitalization = true,
		verbose_errors = true,
		time_report = false
	});
}

/*** Diagnostics. ***/
<location>Stack::t loc_stack;
bool bailout;
location last_location;

section init 
{
	loc_stack = Stack::create();
	last_location = ({
			file = "unknown", 
			first_line = 0,
			first_column = 0,
			last_line = 0,
			last_column = 0
			});
	bailout = false;
}

void message(location loc, string msg)
{
	string beg = loc.file + ":" + 
		     itoa(loc.first_line) + ":" +
		     itoa(loc.first_column);

	if (loc.last_line != 0 &&
	    (loc.last_line != loc.first_line ||
	     (loc.last_column != 0 && loc.last_column != loc.first_column)))
	     	beg = beg + "-" + itoa(loc.last_line) + ":" + 
		      itoa(loc.last_column);
	
	prerr_endline(beg + ": " + msg);
}

void warnl(location loc, string msg)
{
	message(loc, "warning: " + msg);
	if (flags.werror)
		bailout = true;
}

void errorl(location loc, string msg)
{
	message(loc, msg);
	bailout = true;
}

/* Location stack. */
void push_location(location loc)
{
	last_location = loc;
	Stack::push(loc_stack, loc);
}

void pop_location()
{
	Stack::pop(loc_stack);
}

location top_location()
{
	if (Stack::empty(loc_stack))
		return last_location;
	else
		return Stack::top(loc_stack);
}

void warn(string msg)
{
	warnl(top_location(), msg);
}

void error(string msg)
{
	errorl(top_location(), msg);
}

void maybe_choke()
{
	if (bailout == true)
		raise Exit;
}

'a locate(*('a ('b)) f, <'b>located x)
{
	'a r;

	push_location(x.loc);
	try {
		r = f(x.a);
	} finally {
		pop_location();
	}

	return r;
}

'a ice(string msg)
{
	if (bailout == true) {
		error("confused by earlier errors, bailing out [" + msg + "]");
		raise Exit;
	}
	
	error("internal compiler error: " + msg);
	return abort();
}


/*** General utils. ***/
'a get_binding(<'a>bindable sym)
{
	switch sym {
	case Symbol[x]:
		return x;
	case SimpleName[n]:
		return ice("get_binding(): symbol '" + n + "' not bound");
	case QualifiedName[m, n]:
		return ice("get_binding(): qualified symbol '" + 
			   m + "::" + n + "' not bound");
	}
}

void dbg(string s)
{
	print_string(s);
}

bool same_symbols(<'a>symbol s1, <'a>symbol s2)
{
	return s1.module == s2.module && s1.name == s2.name;
}

*('a ('b)) partial(*('a ('c, 'b)) f, 'c c)
{
	return (fun 'a ('b b) (f(c, b)));
}

*('a ('b)) partial2(*('a ('c, 'd, 'b)) f, 'c c, 'd d)
{
	return (fun 'a ('b b) (f(c, d, b)));
}

ty type_of(expr e)
{
	switch e.a.expr_type {
	case Some[t]:
		return t;
	case None:
		return ice("Util::type_of(): untyped expression");
	}
}

ft_record get_ft_record(ty t)
{
	switch t {
	case Func[ft]:
		return ft;
		
	case _:
		return ice("Util::get_ft_record(): non-func type");
	}
}

// Unique name generator

struct ung_counter {
	int cnt;
}

<ung_counter>Hashtbl::t ung_counters;

section init
{
	ung_counters = Hashtbl::create(13);
}

string genname(string prefix)
{
	int cnt;
	
	if (Hashtbl::mem(ung_counters, prefix)) {
		let c = Hashtbl::find(ung_counters, prefix) in
		cnt = c.cnt++;
	} else {
		cnt = 1;
		Hashtbl::add(ung_counters, prefix, ({ cnt = 2 }));
	}

	return prefix + "$" + itoa(cnt);
}

string mangle_type(ty t)
{
    string mangle_symbol(<'a>symbol s)
    {
	return s.module + "::" + s.name;
    }
    
    string mangle_field(<<'a>symbol>field f)
    {
	    return mangle_type(f.fld_type) + 
		   mangle_symbol(get_binding(f.fld_symbol));
    }
    
    string record_content(<<<'a>symbol>field>List::t lst)
    {
	    return String::concat("><", List::map(mangle_field, lst));
    }

    string mangle_list(<ty>List::t l)
    {
    	return String::concat("", List::map(mangle_type, l));
    }

    switch t {
    case Int:		return "i";
    case Bool:		return "b";
    case Void:		return "v";
    case String:	return "s";
    case Float:		return "f";
    case Exn:		return "e";
    case AbstractType:	return "a";
    case Tuple[lst]:
	return "T<" + mangle_list(lst) + ">";
	
    case Func[ft]:
	return "F<" + mangle_type(ft.ret_type) + 
		mangle_list(List::map(fst, ft.parms)) + ">";
		   
    case TyVar[n]:
	return "V<" + n + ">";
	    
    case XTyVar[x]:
    	switch x.t {
	case Some[tt]:
		return mangle_type(tt);
	case None:
		return "X<" + x.name + ">";
	}

    case Struct[lst]:
	    return "S<<" + record_content(lst) + ">>";

    case OptStruct[lst]:
	    return "O<<" + record_content(lst) + ">>";

    case Union[lst]:
	    return "U<<" + record_content(lst) + ">>";
    
    case NamedType[nt]:
	return "N<<" + mangle_list(nt.parms) + "><" + 
	       mangle_symbol(get_binding(nt.tydef)) + ">>";
    }
}
