/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 539 $ $LastChangedDate: 2002-08-25 13:25:13 +0200 (Sun, 25 Aug 2002) $
 *
 * Standard library for Gont, Hashtbl module implementation.
 *
 * Mutable hash table.
 */

open Hashtbl;

struct <'a>entry {
	<*[string, 'a]>List::t lst;
};

struct <'a>t {
	int size;
	int nelem;
	<<*[string, 'a]>List::t>Array::t tab;
};

<'a>t create(int sz)
{
	if (sz < 5)
		sz = 5;

	let ini = fun <*[string, 'a]>List::t (int x) (List::Nil) in
	return ({ nelem = 0,
		  size = sz,
		  tab = Array::init(sz, ini) });
}

<'a>t copy(<'a>t src)
{
	return ({ nelem = src.nelem,
		  size = src.size,
		  tab = Array::copy(src.tab) });
}

int length(<'a>t h)
{
	return h.nelem;
}

void iter(<'a>t h, *(void (string, 'a)) f)
{
	void f2(*[string, 'a] p)
	{
		let [k, v] = p in
		f(k, v);
	}
	
	for (int i = 0; i < h.size; i++)
		List::iter(f2, Array::get(h.tab, i));
}

void check_rehash(<'a>t h)
{
	if (h.nelem < h.size)
		return;
		
	<'a>t n = create(h.nelem * 3);

	let f = (fun void (string k, 'a v) (add(n, k, v))) in
	iter(h, f);

	h.tab = n.tab;
	h.size = n.size;
}

'a find_lst(<*[string, 'a]>List::t lst, string key)
{
	switch lst {
	case List::Nil:
		raise Not_found;
	case List::Cons[[k, v], tl]:
		if (k == key)
			return v;
		else
			return find_lst(tl, key);
	}
}

'a find(<'a>t h, string k)
{
	let e = Array::get(h.tab, String::hash(k) % h.size) in
	return find_lst(e, k);
}

<'a>List::t find_all_lst(<*[string, 'a]>List::t lst, string key)
{
	switch lst {
	case List::Nil:
		return List::Nil;
	case List::Cons[[k, v], tl]:
		if (k == key)
			return List::Cons[v, find_all_lst(tl, key)];
		else
			return find_all_lst(tl, key);
	}
}

<'a>List::t find_all(<'a>t h, string k)
{
	let e = Array::get(h.tab, String::hash(k) % h.size) in
	return find_all_lst(e, k);
}

bool mem(<'a>t h, string k)
{
	bool xfind(<*[string, 'a]>List::t lst)
	{
		switch lst {
		case List::Nil:
			return false;
		case List::Cons[[key, _], tl]:
			if (k == key)
				return true;
			else
				return xfind(tl);
		}
	}
	
	let e = Array::get(h.tab, String::hash(k) % h.size) in
	return xfind(e);
}

void add(<'a>t h, string k, 'a v)
{
	h.nelem++;
	check_rehash(h);
	let pos = String::hash(k) % h.size in
	let e = Array::get(h.tab, pos) in
	Array::set(h.tab, pos, List::Cons[[k, v], e]);
}

void remove(<'a>t h, string k)
{
	<*[string, 'a]>List::t xfind(<*[string, 'a]>List::t lst)
	{
		switch lst {
		case List::Nil:
			raise Not_found;
		case List::Cons[hd, tl]:
			let [key, _] = hd in
			if (k == key)
				return tl;
			else
				return List::Cons[hd, xfind(tl)];
		}
	}
	
	let idx = String::hash(k) % h.size in
	Array::set(h.tab, idx, xfind(Array::get(h.tab, idx)));
}

void remove_all(<'a>t h, string k)
{
	bool is_our_key(*[string, 'a] p)
	{
		let [key, _] = p in return (k == key);
	}
	
	let idx = String::hash(k) % h.size in
	let l = Array::get(h.tab, idx) in
	Array::set(h.tab, idx, List::filter(is_our_key, l));
}
